/**
	Program to calculate the cosine transform of
	an input image

	Fraga, Jul 30, 2001

**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "matrix.h"
#include "tiff.h"

/** This function calculate the Discrete Cosine Transform (DCT) 
	if kind == 0 Direct evaluation
	if kind == 1 Inverse evaluation  **/

static float *table;
static float alpha0, alpha1;
float *buf1, *buf2;

void rescale( float **imgin, byte **imgout, int rows, int cols )
{   
	float fmin, fmax, val, scale;
	int i, j;

	/** We rescale the dinamic range from 0 to 255 **/
	fmin = 1e30;
	fmax = -1e30;
	for ( i=0; i<rows; i++ ){
		for ( j=0; j<cols; j++ ){
			val = imgin[i][j]; 
			if ( fmin > val ) fmin = val;
			if ( fmax < val ) fmax = val;
		}   
	}
	printf( "max = %f\n", fmax );
	printf( "min = %f\n", fmin );

	scale = 255.0/(fmax-fmin);
	for ( i=0; i<rows; i++ ){
		for ( j=0; j<cols; j++ ) 
			imgout[i][j] = (byte)(scale*( imgin[i][j] - fmin ) + 0.5 );
	}
}

void init_dct( int n )
{
	int i, n2;

	n2= n*2;

	table = (float *)malloc( n2 * sizeof(float) );
	if ( table==NULL ) {
		fprintf( stderr, "ERROR allocating memory for coefficients table\n" );
		exit( -1 );
	}
	/** Doing table of coefficients **/
	table[ 0 ] = 1.0;
	for ( i=1; i < n2; i++ )
		table[ i ] = cos ( i * M_PI / n2 ); 

	alpha0 = sqrt( 1.0/n);
	alpha1 = sqrt( 2.0/n );

	buf1 = (float *)malloc( n * sizeof(float) );
	buf2 = (float *)malloc( n * sizeof(float) );
	if ( buf1==NULL || buf2==NULL ) {
		fprintf( stderr, "ERROR allocating memory for coefficients table\n" );
		exit( -1 );
	}
}

void free_dct( )
{
	free( table );
	free( buf1 );
	free( buf2 );
	alpha0 = alpha1 = 0;
}

void dct( float *in, float *out, int n, int kind )
{
	register int u, x, j, k, s;
	int n2, k2;
	float sum;

	n2= 2*n;

	if ( !kind ) {
		sum = 0;
		for( x=0; x<n; x++ ){
			sum += in[x];
		}
		out[0] = alpha0 * sum;

		for( u=1; u<n; u++ ){
			sum = 0;
			for( x=0; x<n; x++ ){
				k = x<<1;
				k = ++k * u;
				j = k % n2;
				s = k / n2;
				if ( s%2 ) j = n2 - j;
				sum += in[x] * table[j];
			}
			out[u] = alpha1 * sum;
		}	
	}
	else {
		for( x=0; x<n; x++ ){
			k = x<<1;
			k2 = ++k;
			sum = in[0] * alpha0;
			for( u=1; u<n; u++ ){
				k = k2 * u;
				j = k % n2;
				s = k / n2;
				if ( s%2 ) j = n2 - j;
				sum += alpha1 * in[u] * table[j];
			}
			out[x] = sum;
		}	
	}
}

MATRIX * image_dct( MATRIX *IN, int kind )
{
	MATRIX *OUT;
	int i, j;
	int n, n2;
	float **imgin, **imgout;

	if ( IN->rows != IN->cols ) {
		fprintf( stderr, "ERROR: Input image must be a square\n" );
		return NULL;
	}
	n = IN->rows;
	n2 = 2*n; 

	OUT = matrix_allocate( n, n, sizeof(float) );
	if ( OUT==NULL ) {
		fprintf( stderr, "ERROR allocating memory for output image\n" );
		return NULL;
	}


	fprintf( stderr, "Hago calculos\n" );
	imgin = (float **)IN->ptr;
	imgout = (float **)OUT->ptr;

	if ( kind == 0 ) {
		/** Column transforms **/
		for ( j=0; j<n; j++ ) {
			for ( i=0; i<n; i++ ) buf1[i] = imgin[i][j];

			dct ( buf1, buf2, n, 0 ); 

			for ( i=0; i<n; i++ ) imgout[i][j] = buf2[i];
		}
		/** Row transforms **/
		for ( i=0; i<n; i++ ) {
			dct ( imgout[i], buf2, n, 0 ); 

			for ( j=0; j<n; j++ ) imgout[i][j] =  buf2[j];
		}
	}
	else {
		/** Row transforms **/
		for ( i=0; i<n; i++ ) 
			dct ( imgin[i], imgout[i], n, 1 ); 
		
		/** Column transforms **/
		for ( j=0; j<n; j++ ) {
			for ( i=0; i<n; i++ ) buf1[i] = imgout[i][j];

			dct ( buf1, buf2, n, 1 ); 

			for ( i=0; i<n; i++ ) imgout[i][j] = buf2[i];
		}
	}
	return OUT;	
}


int do_work( char *namein ) 
{
	MATRIX *IN, *M, *OUT, *T;
	FILEINFO finfo;
	int i, j;
	byte *iptr;
	float *fptr, *fptr2;

#ifdef TEST
	float kk[8] = {1,1,1,1,0,0,0,0};
	float r[8], r2[8];

	init_dct(8);
	printf( "# Sequence     Direct DCT    Inverse DCT\n" );

	dct( kk, r, 8, 0);
	dct( r, r2, 8, 1);

	for ( i=0; i<8; i++ ) printf( "%d %e %e %e\n", i,  kk[i], r[i], r2[i] );
	free_dct( );
#endif
	
	IN = read_tiff( &finfo, namein ); 
	if ( IN==NULL ) {
		fprintf( stderr, "ERROR reading %s file\n", namein );
		return -1;
	}
	if ( IN->type != 5 ) {
		fprintf( stderr, "ERROR Bad format. Image must be in 256 grey tones\n");
		return -1;
	}
	if ( IN->rows != IN->cols ) {
		fprintf( stderr, "ERROR: Input image must be a square\n");
		return -1;
	}
	M = matrix_allocate( IN->rows, IN->cols, sizeof(float) );
	if ( M==NULL ) {
		fprintf( stderr, "ERROR allocating memory for output image\n" );
		return -1;
	}
	/** We copy the input image to a float image buffer **/
	for ( i=0; i<IN->rows; i++ ) {
		iptr = (byte *) IN->ptr[i];
		fptr = (float *) M->ptr[i];
		for ( j=0; j<IN->cols; j++ ) *fptr++ = *iptr++;
	}
	/** We initialize table and buffers for calculate the DCT **/
	init_dct( IN->rows );

	OUT = image_dct( M, 0 );

	for ( i=0; i<IN->rows; i++ ) {
		fptr = (float *) OUT->ptr[i];
		fptr2 = (float *) M->ptr[i];
		for ( j=0; j<IN->cols; j++ ) *fptr2++ = fabs(*fptr++) + 1.0;
	}
	rescale( (float **)M->ptr, (byte **)IN->ptr, IN->rows, IN->cols );
	IN->type=5;
	if ( write_tiff( IN, "logcosine.tif" ) != GOOD_WRITE )  
		return -1;

	T = image_dct( OUT, 1 );
	free_dct( );


	for ( i=0; i<IN->rows; i++ ) {
		iptr = (byte *) IN->ptr[i];
		fptr = (float *) T->ptr[i];
		for ( j=0; j<IN->cols; j++ ) *iptr++ = (byte) ( *fptr++ + 0.2 );
	}

	IN->type=5;
	if ( write_tiff( IN, "test.tif" ) != GOOD_WRITE )  
		return -1;

	matrix_free( T );
	matrix_free( IN );
	matrix_free( M );
	matrix_free( OUT );

	return 0;
}

int main( int argc, char *argv[] )
{
	char namein[32];

	if ( argc < 2 ) { 
		fprintf( stderr, "Use: %s image.tif\n", argv[0] );
		return -1;
	}

	strncpy( namein, argv[1], 32 );

	if ( do_work( namein ) >= 0 ) {
		fprintf( stderr, "OK!!!\n" );
		return 0;
	}
	else {
		fprintf( stderr, "ERROR: Something is wrong!!!\n" );
		exit(1);
	}
	return 0;
}
