/**
	Program to calculate the cosine transform of
	an input image

	Fraga, Jul 30, 2001

**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "matrix.h"
#include "tiff.h"

/** This function calculate the Discrete Cosine Transform (DCT) 
	if kind == 0 Direct evaluation
	if kind == 1 Inverse evaluation  **/
MATRIX * dct( MATRIX *IN, int kind )
{
	MATRIX *OUT;
	int u, v, x, y;
	int n, n2, i, s, k;
	float *ptrin, *ptrout;
	float *table;
	float alpha0, alpha1;
	float au, av, cu, cv, sum;

	if ( IN->rows != IN->cols ) {
		fprintf( stderr, "ERROR: Input image must be a square\n" );
		return NULL;
	}
	n = IN->rows;
	n2 = 2*n; 
	table = (float *)malloc( n2 * sizeof(float) );
	if ( table==NULL ) {
		fprintf( stderr, "ERROR allocating memory for coefficients table\n" );
		return NULL;
	}
	OUT = matrix_allocate( n, n, sizeof(float) );
	if ( OUT==NULL ) {
		fprintf( stderr, "ERROR allocating memory for output image\n" );
		return NULL;
	}
	/** Doing table of coefficients **/
	table[ 0 ] = 1.0;
	for ( i=1; i < n2; i++ )
		table[ i ] = cos ( i * M_PI / n2 ); 

	alpha0 = sqrt( 1.0/n);
	alpha1 = sqrt( 2.0/n );

	fprintf( stderr, "Hago calculos\n" );
	if ( kind == 0 ) {
		for ( u=0; u<n; u++ ) {
			au = ( u==0 ) ? alpha0 : alpha1;	

			ptrout = (float *) OUT->ptr[u];
			for ( v=0; v<n; v++ ) {
				av = ( v==0 ) ? alpha0 : alpha1;	

				sum = 0;
				for ( x=0; x<n; x++ ) {
					ptrin = (float *) IN->ptr[x];
					k = (x*2 + 1) * u;
					i = k % n2;
					s = k / n2;
					if ( s%2 ) i = n2 - i;
					cu = table[i];
					for ( y=0; y<n; y++ ) {
						k = (y*2 + 1) * v;
						i = k % n2;
						s = k / n2;
						if ( s%2 ) i = n2 - i;
						cv = table[i];
						sum += *ptrin++ * cv; 
					}
					sum *= cu;
				}
				*ptrout++ = au * av * sum;
			}
			fprintf( stderr, "Hice renglon %d\n", u );
		}
	}
	else {
		for ( x=0; x<n; x++ ) {
			ptrout = (float *) OUT->ptr[u];
			for ( y=0; y<n; y++ ) {
				sum = 0;
				for ( u=0; u<n; u++ ) {
					ptrin = (float *) IN->ptr[u];
					au = ( u==0 ) ? alpha0 : alpha1;	
					k = (x*2 + 1) * u;
					i = k % n2;
					s = k / n2;
					if ( s%2 ) i = n2 - i;
					cu = table[i];
					for ( v=0; v<n; v++ ) {
						av = ( v==0 ) ? alpha0 : alpha1;	
						k = (y*2 + 1) * v;
						i = k % n2;
						s = k / n2;
						if ( s%2 ) i = n2 - i;
						cv = table[i];

						sum += av * *ptrin++ * cv;
					}
					sum *= au * cu;
				}
				*ptrout ++ = sum;
			}
			fprintf( stderr, "renglon %d\n", x );
		}
	}
	free ( table );
	return OUT;	
}


int do_work( char *namein ) 
{
	MATRIX *IN, *M, *OUT, *T;
	FILEINFO finfo;
	int i, j;
	byte *iptr;
	float *fptr;
	
	IN = read_tiff( &finfo, namein ); 
	if ( IN==NULL ) {
		fprintf( stderr, "ERROR reading %s file\n", namein );
		return -1;
	}
	if ( IN->type != 5 ) {
		fprintf( stderr, "ERROR Bad format. Image must be in 256 grey tones\n");
		return -1;
	}
	if ( IN->rows != IN->cols ) {
		fprintf( stderr, "ERROR: Input image must be a square\n");
		return -1;
	}
	M = matrix_allocate( IN->rows, IN->cols, sizeof(float) );
	if ( M==NULL ) {
		fprintf( stderr, "ERROR allocating memory for output image\n" );
		return -1;
	}
	/** We copy the input image to a float image buffer **/
	for ( i=0; i<IN->rows; i++ ) {
		iptr = (byte *) IN->ptr[i];
		fptr = (float *) M->ptr[i];
		for ( j=0; j<IN->cols; j++ ) *fptr++ = *iptr++;
	}
	fprintf( stderr, "Hare dct\n" );
	OUT = dct( M, 0 );
	fprintf( stderr, "Hice dct\n" );

	T = dct( OUT, 1 );

	for ( i=0; i<IN->rows; i++ ) {
		iptr = (byte *) IN->ptr[i];
		fptr = (float *) T->ptr[i];
		for ( j=0; j<IN->cols; j++ ) *iptr++ = (byte) *fptr++;
	}

	IN->type=5;
	if ( write_tiff( IN, "otra.tif" ) != GOOD_WRITE )  
        return -1;
	matrix_free( IN );
	matrix_free( M );
	matrix_free( OUT );

	return 0;
}

int main( int argc, char *argv[] )
{
	char namein[32];

	if ( argc < 2 ) { 
		fprintf( stderr, "Use: %s image.tif\n", argv[0] );
		return -1;
	}

	strncpy( namein, argv[1], 32 );

	if ( do_work( namein ) >= 0 ) {
		fprintf( stderr, "OK!!!\n" );
		return 0;
	}
	else {
		fprintf( stderr, "ERROR: Something is wrong!!!\n" );
		exit(1);
	}
	return 0;
}
