/**
	Program to apply homomorphic filtering

	Fraga. Jun 24, 2001
**/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "matrix.h"
#include "tiff.h"
#include "fft.h"

void normalizationf( MATRIX *IN, int rows, int cols )
{
	int i, j;
	float *ptr, imin, imax;
	float scale=0.0;

	imin = 1e36;
	imax = -1e36;
	for( i=0; i<rows; i++ ) {
		ptr = (float *)IN->ptr[i];
		for( j=0; j<cols; j++ ) {
			imin = (imin > ptr[j]) ? ptr[j] : imin;
			imax = (imax < ptr[j]) ? ptr[j] : imax;
		}
	}
	fprintf( stderr, "%e %e\n", imin, imax );
	if ( imin != imax ) {
		scale = 255.0/(imax-imin);
	}
	else {
		fprintf( stderr, "NO hay valores!!!\n" );
	}
//	for( i=0; i<rows; i++ ) {
//		ptr = (float *)IN->ptr[i];
//		for( j=0; j<cols; j++ ) 
//			*ptr = (byte)(scale*(*ptr++ - imin) + 0.5 );
//	}
}

void homo_filter( float **img, int rows, int cols, float r, float yl, float yh )
{
	int i, j;
	int c2=cols/2;
	int c, r_1, r_12;
	float cx, cy, x, y, val, y2, h1, h2, sigma;

	if ( r < 1.0) r *= c2;

	c = rows/10;
	r_1 = r - c;
	/** r_1 = ( r_1 <= 0 ) ? 0 : r_1; **/
	r_12 = r_1 * r_1;

	cy = (rows-1)/2.0;
	cx = (cols-1)/2.0;
	
	sigma = c*1.20112240878644979487;

	for (i=0; i<rows ; i++) {
		y = i - cx;
		y2 = y*y;
		for (j=0; j<=c2;  j++) {
			x = j - cx;
			if ( x*x + y2 < r_12 ) {
				img[i][2*j] *= yl; 
				img[i][2*j+1] *= yl;
			}
			else {
				h1 = y - r_1; 
				h2 = x - r_1; 
				val = h1*h1 + h2*h2;
				val /= sigma*sigma;
				val = yh*(1.0 - exp (-val) ) + yl;
				img[i][2*j] *= val; 
				img[i][2*j+1] *= val;
			}
		}
	}
}

int do_work( char *namein, float r, float yL, float yH ) 
{
	MATRIX *IN;
	MATRIX *FFTIMG;
	FILEINFO finfo;
	int i, j, signo;
	byte *iptr;
	float *fptr;
	

	IN = read_tiff( &finfo, namein ); 
	if ( IN==NULL ) {
		fprintf( stderr, "ERROR reading %s file\n", namein );
		return -1;
	}
	FFTIMG = matrix_allocate( IN->rows, IN->cols + 2, sizeof(float) );

	/** We copy the readed image to the new float matrix **/
	/** The variable signo is for move the FT a half of a period **/
	signo = -1;
	for ( i=0; i<IN->rows; i++ ) {
	 	iptr = (byte *)IN->ptr[i];
	 	fptr = (float *)FFTIMG->ptr[i];
	 	for ( j=0; j<IN->cols; j++ ) {
	 		signo *= -1;
			*fptr++ = (float)log( *iptr + 1 ) * signo;
			iptr++;
	 	}
	 	signo *= -1;
	}

	/** Now we calculate the FFT **/
	image_fft ( (float **)FFTIMG->ptr, IN->rows , IN->cols, DIRECT );

	/** We apply the filter **/
	homo_filter( (float **)FFTIMG->ptr, IN->rows , IN->cols, r, yL, yH );

	/** For testing, we apply the inverse transform **/
	image_fft ( (float **)FFTIMG->ptr, IN->rows , IN->cols, INVERSE );

	/** Reverse the signs **/
	signo = -1;
	for ( i=0; i<IN->rows; i++ ) {
	 	fptr = (float *)FFTIMG->ptr[i];
	 	for ( j=0; j<IN->cols; j++ ) {
	 		signo *= -1;
	 		*fptr *= signo;
			*fptr++ = exp( *fptr );
	 	}
		signo *= -1;
	}
	normalizationf( FFTIMG, IN->rows, IN->cols );

	/** We put the result in a 8 bit matrix and then we store
		the resulted image. This must be the equal to the original image **/
	for ( i=0; i<IN->rows; i++ ){
		iptr = (byte *)IN->ptr[i];
		fptr = (float *)FFTIMG->ptr[i];
		for ( j=0; j<IN->rows; j++ ){
			*iptr++ = (*fptr < 0) ? 0 : (*fptr > 255) ? 255 : *fptr;
			fptr++;
		}
	}
	if ( write_tiff( IN, "homofiltered.tif" ) != GOOD_WRITE )
		return -1;

	matrix_free( IN );
	matrix_free( FFTIMG );

	return 0;
}

int main( int argc, char *argv[] )
{
	char namein[32];
	float radius, yh, yl;

	yl = 10;
	yh = log( yl );
	printf( "%e %e\n", yl, yh );

	if ( argc < 5 ) { 
		fprintf( stderr, "Use: %s image.tif radius yL yH\n", argv[0] );
		return -1;
	}

	strncpy( namein, argv[1], 32 );
	radius = atof( argv[2] );
	yl = atof( argv[3] );
	yh = atof( argv[4] );

	fprintf( stderr, "%s %f %f %f\n", namein, radius, yl, yh );

	if ( do_work( namein, radius, yl, yh ) >= 0 ) {
		fprintf( stderr, "OK!!!\n" );
		return 0;
	}
	else {
		fprintf( stderr, "ERROR: Something is wrong!!!\n" );
		exit(1);
	}
	return 0;
}
